//
// support.c
//
// lwIP/VDK implementation of the various support functions required by
// the common ADI-TTCP code.
//

#include "config.h"
#include "adi_ttcp.h"
#include "getopt.h"
#include "Vdk.h"

/////////////////////////////////////////////////////////////////////////////
//// STATIC VARIABLES
//

static ker_sem_id_t transfer_terminated;

volatile ULONGLONG num_idle_loops;
double             idle_loop_time;	// time in seconds for one idle loop


//
// return pointer to a static struct tm with the hour, min and sec fields
// initialised from the millisecond count t
//
static struct tm* elapsedtime(TIME t)
{
    int              s = (t+500) / 1000;
    static struct tm r = {0};

    r.tm_hour = (s / 3600);
    s -= (r.tm_hour * 3600);
    r.tm_min  = (s / 60);
    s -= (r.tm_min * 60);
    r.tm_sec = s;

    return (&r);

}

//
// Print a time-stamped message to stdout
//
void TTCP_ConsoleMsg( const char *format, ... )
{
    FILE *logfd = stdout;
    TIME elapsed_ms = TTCP_GetMillisecCount();
    struct tm* t = elapsedtime(elapsed_ms);
    char  szBuffer[ 400 ];
    size_t ts_len, msg_len;
    char* p;
    va_list  marker;

    // insert timestamp into buffer
    sprintf(szBuffer, "[%02d:%02d:%02d] ", t->tm_hour, t->tm_min, t->tm_sec);
    ts_len = strlen(szBuffer);

    // append caller's message
    va_start( marker, format );
    vsprintf( szBuffer+ts_len, format, marker );
    va_end(marker);

    // print each line and follow with ts_len spaces if not final line
    msg_len = strlen(szBuffer);
    p = szBuffer;
    while (1)
    {
	char* q = p;

	p = strchr(p, '\n');
	if (p && (&szBuffer[msg_len] - p) > 1)
	{
	    fprintf(logfd, "%.*s%*s", p-q+1, q, ts_len, " ");
	    p += 1;
	}
	else
	{
	    fprintf(logfd, "%s", q);
	    break;
	}
    }
}

//
// return the last lwIP error
//
DWORD TTCP_GetLastError()
{
    return ( ker_get_lasterror() );
}

// return relative time measured in milliseconds
TIME TTCP_GetMillisecCount(void)
{
    TIME now = clock();
    return ( (now - TTCP_start_time) / (CLOCKS_PER_SEC/1000) );
}

// sleep for at least the requested number of milliseconds
void TTCP_SleepMillisec(DWORD millisec)
{
    ker_sleep(millisec);
}

// delay for at least the requested number of microseconds
extern void TTCP_DelayMicrosec(DWORD microsec)
{
    clock_t delay = (CLOCKS_PER_SEC/1000000) * microsec;
    clock_t t1, t2;

    t1 = clock();
    t2 = clock();
    while ((t2 - t1) < delay)
	t2 = clock();
}

//
// For each transfer in tfr_list create a thread that calls dispatch() passing
// in a pointer to the transer's CMD_BLOCK. Then wait for all threads to finish.
//
// Return 0 if all threads ran OK or non-zero otherwise.
//
int TTCP_RunTransfers(PTEST_BLOCK tfr_list, int (dispatch)(void*))
{
    int transfer_count = 0;
    int r;

    // get a semaphore for threads to use to report termination
    transfer_terminated = ker_get_semaphore( 0, UINT_MAX );

    // create and start a thread for each transfer
    while (tfr_list)
    {
	ker_thread_id_t th_id;
	PTEST_BLOCK     ptfr = tfr_list;

	tfr_list = ptfr->m_chain;

	// specify type of thread to be created
	ker_set_auxdata( (void*)kADI_TTCP_ThreadType );

	r = ker_create_thread(
	          (ThreadEntryPoint)dispatch,	        // entrypoint ...
				    &ptfr->m_Cmd,       // ... and argument
				    0,		        // default stack size
	                            NULL,               // default stack area
				    ker_kPriorityLevel4,// priority
				    "",                 // name
				    &th_id              // thread ID
	                        );

	if ( r == ker_err_thread_cre )
	{
	    TTCP_ConsoleMsg( "ERROR: failed to create thread for TRANSFER %d\n",
	                    ptfr->m_id );
	    exit( 5 );
	}

	transfer_count += 1;
    }

    // wait for all the threads to finish, with no timeout
    while (transfer_count)
    {
	ker_pend_semaphore( transfer_terminated, 0 );
	transfer_count -= 1;
    }

    // free up the semaphore
    ker_rel_semaphore( transfer_terminated );

    // return 0 if all threads terminated nicely, otherwise 1
    return (0); // be optimistic
}

// terminate the running thread
void TTCP_TerminateThread(PTEST_BLOCK ptfr)
{
#if 0 // takes too long over standard USB / JTAG connection
    TTCP_ConsoleMsg( "TRANSFER %d is terminating\r\n", ptfr->m_id);
#endif
    ker_post_semaphore( transfer_terminated );

    ker_destroy_thread( ker_get_current_thread() );
}

// create and run a thread to perform a RELAY command; return immediately
extern void TTCP_CreateRelay(PRELAY_BLOCK prly, void (relay)(void*))
{
    ker_thread_id_t th_id;
    int             r;

    // specify type of thread to be created
    ker_set_auxdata( (void*)kADI_TTCP_ThreadType );

    r = ker_create_thread(
	    (ThreadEntryPoint)relay,	          // entrypoint ...
			      prly,               // ... and argument
			      0,		  // default stack size
			      NULL,               // default stack area
			      ker_kPriorityLevel7,// priority
			      "",                 // name
			      &th_id              // thread ID
			);

    if ( r == ker_err_thread_cre )
    {
	TTCP_ConsoleMsg( "ERROR: failed to create RELAY thread\n" );
	exit( 6 );
    }

}

// perform any target specific start-of-transfer actions
void TTCP_TransferStarting(PTEST_BLOCK ptfr)
{
    // record initial value of idle loop counter
    ptfr->m_num_idle_loops = num_idle_loops;

}

// perform any target specific end-of-transfer actions
void TTCP_TransferStopping(PTEST_BLOCK ptfr)
{
    // calculate number of idle loop executions during this transfer
    ptfr->m_num_idle_loops = num_idle_loops - ptfr->m_num_idle_loops;
}

// return CPU idle time (in seconds) for this transfer, if available
double TTCP_CPUIdleTime(PTEST_BLOCK ptfr)
{
    return (ptfr->m_num_idle_loops * idle_loop_time);
}

static void innerIdleFunction(int prm)
{
    volatile int v1 = 0;
    volatile int v2 = prm;

    while (v2 > 0)
    {
	v1++;
	v2--;
    }
}

void TTCP_IdleFunction()
{
    int j;

    for (j = 0; j < 250; j ++)
	innerIdleFunction(j);

    num_idle_loops += 1;
}

void TTCP_CalibrateIdleFunction()
{
    int i;
    TIME t1, t2;

    TTCP_ConsoleMsg("Calibrating idle loop\n" );

    ker_disable_scheduler();
    t1 = clock();
    for (i = 0; i < 1000; i += 1)
    {
	TTCP_IdleFunction();
    }
    t2 = clock();
    ker_enable_scheduler();

    idle_loop_time = (((double)(t2 - t1) / CLOCKS_PER_SEC) / 1000.0);
}

