#include "config.h"
#include "adi_ttcp.h"
#include "getopt.h"

#include <iphlpapi.h>

/////////////////////////////////////////////////////////////////////////////
//// GLOBAL VARIABLES
//

extern int errno;

BOOL     g_bShutdown = FALSE;               // Set by Ctrl-C Handler
WSADATA  g_WsaData;

LARGE_INTEGER gPerformanceFreq;

//
// Print a simple usage message to stderr
//
static void usage(const char* prog)
{
    fprintf(stderr,
            "Usage: %s [-p port]\n"
	    "       port is port to listen on for control stream connection.\n"
	    "       Default port is %d.\n",
            prog, CONTROL_PORT);
}

//
// Print a time-stamped message to stderr
//
void TTCP_ConsoleMsg( const char *format, ... )
{
    FILE *logfd = stderr;
    time_t now = time(NULL);
    struct tm* t = localtime(&now);
    char  szBuffer[ 400 ];
    size_t ts_len, msg_len;
    char* p;
    va_list  marker;

    // insert timestamp into buffer
    sprintf(szBuffer, "[%02d:%02d:%02d] ", t->tm_hour, t->tm_min, t->tm_sec);
    ts_len = strlen(szBuffer);

    // append caller's message
    va_start( marker, format );
    vsprintf( szBuffer+ts_len, format, marker );
    va_end(marker);

    // print each line and follow with ts_len spaces if not final line
    msg_len = strlen(szBuffer);
    p = szBuffer;
    while (1)
    {
	char* q = p;

	p = strchr(p, '\n');
	if (p && (&szBuffer[msg_len] - p) > 1)
	{
	    fprintf(logfd, "%.*s%*s", p-q+1, q, ts_len, " ");
	    p += 1;
	}
	else
	{
	    fprintf(logfd, "%s", q);
	    break;
	}
    }
}

//
// return the last Winsock 2 error
//
DWORD TTCP_GetLastError()
{
    return ( WSAGetLastError() );
}

// return relative time measured in milliseconds
TIME TTCP_GetMillisecCount(void)
{
    return (GetTickCount());
}

// sleep for at least the requested number of milliseconds
void TTCP_SleepMillisec(DWORD millisec)
{
    Sleep(millisec);
}

// delay for at least the requested number of microseconds
extern void TTCP_DelayMicrosec(DWORD microsec)
{
    LARGE_INTEGER t1, t2;

    QueryPerformanceCounter(&t1);
    QueryPerformanceCounter(&t2);
    while ((((t2.QuadPart - t1.QuadPart) * 1000000) / gPerformanceFreq.QuadPart)
 	        < microsec)
	QueryPerformanceCounter(&t2);
}

//
// For each transfer in tfr_list create a thread that calls dispatch() passing
// in a pointer to the transer's CMD_BLOCK. Then wait for all threads to finish.
//
// Return 0 if all threads ran OK or non-zero otherwise.
//
int TTCP_RunTransfers(PTEST_BLOCK tfr_list, int (dispatch)(void*))
{
    // array of HANDLE for use in WaitForMultipleObjects()
    HANDLE threads[128];
    DWORD  num_thread = 0;

    DWORD  r;

    // create and start a thread for each transfer
    while (tfr_list)
    {
	PTEST_BLOCK ptfr = tfr_list;

	tfr_list = ptfr->m_chain;

	threads[num_thread++] = CreateThread(
	                            NULL,	  // default security attributes
				    0,		  // default stack size
	    (LPTHREAD_START_ROUTINE)dispatch,	  // entrypoint
	                            &ptfr->m_Cmd, // argument
				    0,		  // create-and-go
	                            NULL          // don't need thread ID
	                        );

	if (threads[num_thread-1] == NULL)
	{
	    TTCP_ConsoleMsg("ERROR: failed to create thread for TRANSFER %d\n",
	                    ptfr->m_id);
	    ExitProcess(5);
	}
    }

    // wait for all the threads to finish, with no timeout
    r = WaitForMultipleObjects(num_thread, threads, TRUE, INFINITE);

    // close all thread handles
    while (num_thread--)
    {
	CloseHandle(threads[num_thread]);
    }

    // return 0 if all threads terminated nicely, otherwise 1
    return (((r >= WAIT_OBJECT_0) && (r < WAIT_ABANDONED_0))? 0 : 1);
}

// terminate the running thread because of an error
void TTCP_TerminateThread(PTEST_BLOCK ptfr)
{
    //TTCP_ConsoleMsg( "TRANSFER %d is terminating\r\n", ptfr->m_id);

    ExitThread(1);
}

// create and run a thread to perform a RELAY command; return immediately
extern void TTCP_CreateRelay(PRELAY_BLOCK prly, void (relay)(void*))
{
    HANDLE h;

    h = CreateThread(
			NULL,	// default security attributes
			0,	// default stack size
(LPTHREAD_START_ROUTINE)relay,	// entrypoint
			prly,	// argument
			0,	// create-and-go
			NULL	// don't need thread ID
	);

    if (h == NULL)
    {
	TTCP_ConsoleMsg("ERROR: failed to create RELAY thread\n");
	ExitProcess(6);
    }

    CloseHandle(h);
}

// perform any target specific start-of-transfer actions
void TTCP_TransferStarting(PTEST_BLOCK ptfr)
{
    return;
}

// perform any target specific end-of-transfer actions
void TTCP_TransferStopping(PTEST_BLOCK ptfr)
{
    return;
}

// return CPU idle time (in seconds) for this transfer, if available
double TTCP_CPUIdleTime(PTEST_BLOCK ptfr)
{
    return (-1.0);  // not available
}


//
// Console Ctrl-C handler
//
BOOL WINAPI CtrlHandler( DWORD dwCtrlType )
{
   switch( dwCtrlType )
   {
      case CTRL_C_EVENT:
         TTCP_ConsoleMsg( "Ctrl+C Event\n"
                          "ADI-TTCP finished\n" );
         g_bShutdown = TRUE;

         //
         // Flush Logged Messages
         // ---------------------
         // This allows redirected console output to be saved OK.
         //
         fflush( stderr );
         fflush( stdout );

         ExitProcess(4);
         return( TRUE );

      case CTRL_BREAK_EVENT:
      case CTRL_CLOSE_EVENT:
      case CTRL_SHUTDOWN_EVENT:
      default:
         return( FALSE );
   }
}


/////////////////////////////////////////////////////////////////////////////
//// MAIN Program Entry Point
//

/////////////////////////////////////////////////////////////////////////////
//// main()
//
// Purpose
// Console application main program entry point. Starts up winsock2 and then
// enters main loop awaiting control stream connection.
//
// Parameters
//   [-p port]  where port is the port number on which to listen for control
//              stream connections
//
// Return Value
//   returns zero(0) for normal return
//
int main( int argc, char **argv )
{
    int r;

    //
    // Check for non-default control stream port
    //
    if (argc > 1)
    {
	int c;

	while ( optind != argc )
	{
	    c = getopt(argc, argv, "p:");

	    switch (c)
	    {
	      case EOF:
		optarg = argv[optind];
		optind++;
		break;

	      case 'p':
		control_port = atoi(optarg);
	        break;

	      default:
		usage(argv[0]);
	        return (1);
	    }
        }

    }

    //
    // Say Hello
    //
    TTCP_ConsoleMsg( "ADI-TTCP rev %s started\n", ADI_TTCP_VERSION);

    //
    // Start Winsock 2
    //
    if( WSAStartup( MAKEWORD(0x02,0x00), &g_WsaData ) == SOCKET_ERROR )
    {
	int nError = WSAGetLastError();

	TTCP_ConsoleMsg( "*** Winsock Error: %s Failed; ERROR: %d (0x%8.8X)\n",
			 "WSAStartup", nError, nError );
	return( 3 );
    }

    //
    // Set The Ctrl-C Handler
    //
    SetConsoleCtrlHandler( CtrlHandler, TRUE );

    // find out processor speed
    QueryPerformanceFrequency(&gPerformanceFreq);

    // find out and print hostname and network adapter information
    {
	PIP_ADAPTER_INFO pAdapterInfo  = NULL;
	PIP_ADAPTER_INFO pAdapter      = NULL;
	ULONG            ulOutBufLen   = sizeof(IP_ADAPTER_INFO);
	DWORD            dwRetVal      = 0;

	char*            pbuf          = (char*)malloc(8192);
	char*            p             = pbuf;
	int              n             = 0;

	p += sprintf(pbuf, "Host Name: ");
	gethostname(p, 1024);
	p = pbuf + strlen(pbuf);

	pAdapterInfo = (IP_ADAPTER_INFO *) malloc( sizeof(IP_ADAPTER_INFO) );

	// Make an initial call to GetAdaptersInfo to get
	// the necessary size into the ulOutBufLen variable
	if (GetAdaptersInfo( pAdapterInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) {
	    free(pAdapterInfo);
	    pAdapterInfo = (IP_ADAPTER_INFO *) malloc (ulOutBufLen); 
	}

	if ((dwRetVal = GetAdaptersInfo( pAdapterInfo, &ulOutBufLen)) == NO_ERROR) {
	    pAdapter = pAdapterInfo;
	    while (pAdapter) {
		PIP_ADDR_STRING pip = &pAdapter->IpAddressList;

		p += sprintf(p, "\nNetwork Adapter %d:", n++);
		while (pip) {
		    p += sprintf(p, " %s", pip->IpAddress.String);
		    pip = pip->Next;
		}

		pAdapter = pAdapter->Next;
	    }

	    TTCP_ConsoleMsg("%s\n", pbuf);
	}
	else {
	    printf("Call to GetAdaptersInfo failed.\n");
	}
	free(pAdapterInfo);
	free(pbuf);
    }

    // perform main loop
    r = TTCP_MainLoop();

    // tidy up
    WSACleanup();

    // Say Goodbye
    TTCP_ConsoleMsg( "ADI-TTCP finished\n");

    return( r );
}

